"""Base class of tokenizer"""
from abc import ABC, abstractmethod
from typing import List, Tuple

from .tasks.normalizer.mappings import MappingDict


class Tokenizer(ABC):
    """
    Abstract class for tokenizing
    """

    def __init__(self):
        self._en_mapping = MappingDict.load_jsons(["digit_en", "punc_en"])
        self._space_mapping = MappingDict.load_jsons(["space_keep"])

    def word_tokenize(self, text: str) -> List[str]:
        """
        Tokenize the input text into a list of words.
        :param text: The input text to tokenize.
        :return: A list of tokenized words.
        """
        tokens = self.word_span_tokenize(text)
        return [text for (_, _, text) in tokens]

    @abstractmethod
    def word_span_tokenize(self, text: str) -> List[Tuple[int, int, str]]:
        """
        Tokenize the input text and return spans of the tokenized words.
        :param text: The input text to tokenize.
        :return: A list of tuples containing the start index, end index, and the tokenized word for each word span.
        """

    def sentence_tokenize(self, text: str) -> List[str]:
        """
        Tokenize the input text into a list of sentences.
        :param text: The input text to tokenize.
        :return: A list of sentences.
        """
        tokens = self.sentence_span_tokenize(text)
        return [text for (_, _, text) in tokens]

    @abstractmethod
    def sentence_span_tokenize(self, text: str, clean_before_tokenize: bool = True) -> List[Tuple[int, int, str]]:
        """
        Tokenize the input text and return spans of the tokenized sentences.
        :param text: The input text to tokenize.
        :param clean_before_tokenize: clean and then tokenize it
        :return: A list of tuples containing the start index, end index, and the sentence for each sentence span.
        """

    def paragraph_tokenize(self, text: str) -> List[str]:
        """
        Tokenize the input text into a list of paragraph.
        :param text: The input text to tokenize.
        :return: A list of paragraph.
        """
        tokens = self.paragraph_span_tokenize(text)
        return [text for (_, _, text) in tokens]

    def paragraph_span_tokenize(self, text: str) -> List[Tuple[int, int, str]]:
        """
        Tokenize the input text and return spans of the tokenized paragraph.
        :param text: The input text to tokenize.
        :return: A list of tuples containing the start index, end index, and the sentence for each sentence span.
        """
        text2 = self._clean_text(text)
        text2_len = len(text2)
        sentences = self.sentence_span_tokenize(text2, False)
        paragraphs: List[Tuple[int, int, str]] = []
        last_index = 0
        for _, sentence_end, _ in sentences:
            if last_index + 1 >= text2_len:
                break
            pointer = sentence_end
            while True:
                if pointer + 1 >= text2_len:
                    paragraphs.append((last_index, pointer, text[last_index:pointer]))
                    break
                character = text2[pointer]
                if character == "\n":
                    paragraphs.append((last_index, pointer, text[last_index:pointer]))
                    last_index = pointer + 1
                    break
                if character not in self._space_mapping or not self._space_mapping.get(character).is_space:
                    break
                pointer = pointer + 1
        return paragraphs

    def _clean_text(self, text: str) -> str:
        """
        Clean the input text by replacing digits and punctuation with normalized versions.
        :param text: He inputs text to clean.
        :return: The cleaned text with normalized digits and punctuation.
        """
        return ''.join(
            [char if not self._en_mapping.get(char)
             else self._en_mapping.get(char).char for char in text])



"""This module includes a Tokenizer class for tokenizing texts"""
from abc import ABC
from typing import List, Tuple

from spacy.lang.en import English
from spacy.pipeline import Sentencizer

from ...tokenizer import Tokenizer


class SpacyTokenizer(Tokenizer, ABC):
    """
    A class impl tokenizer with spacy
    ...
    Methods
    -------
    word_tokenize(text: str):
        return tokenized text
    sentence_tokenize(text: str):
        return sentence tokenized text
    """

    def __init__(self):
        """
        constructor
        """
        super().__init__()
        self.__nlp = English()
        self.__tokenizer = self.__nlp.tokenizer
        self.__sentencizer = Sentencizer()

    def word_span_tokenize(self, text) -> List[Tuple[int, int, str]]:
        text2 = self._clean_text(text)
        spans = self.__tokenizer(text2)
        return [(span.idx, span.idx + len(span.text), text[span.idx: span.idx + len(span.text)]) for span in spans]

    def sentence_span_tokenize(self, text, clean_before_tokenize=True) -> List[Tuple[int, int, str]]:
        text2 = self._clean_text(text) if clean_before_tokenize else text
        spans = self.__sentencizer(self.__nlp(text2))
        result = []
        last_index = 0
        for span in spans:
            if span.is_sent_end:
                result.append((last_index, span.idx + len(span.text), text[last_index:span.idx + len(span.text)]))
                last_index = span.idx + len(span.text)
        return result



"""This module includes Normalizer class for normalizing texts"""

from abc import ABC, abstractmethod
from typing import List, Tuple


class Normalizer(ABC):
    """
    The Normalizer class is an abstract base class that defines the interface for text normalization. It provides two
    abstract methods: normalize() and span_normalize(), subclasses can implement which to perform specific
    normalization tasks.

    Example Usage

    # Create a subclass of Normalizer
    class MyNormalizer(Normalizer):

        Def normalize (self, text: str) -> str:
            # Implement the normalization logic here
            ...

        Def span_normalize(self, text: str) -> List[Tuple[int, int, str]]:
            # Implement the normalization logic and return spans of normalized tokens
            ...

    # Create an instance of the subclass
    normalizer = MyNormalizer()

    # Normalize a text
    normalized_text = normalizer.normalize("This is a sample text.")

    # Normalize a text and get spans of normalized tokens
    normalized_spans = normalizer.span_normalize("This is a sample text.")
    """

    @abstractmethod
    def normalize(self, text: str) -> str:
        """
        Normalize the input text.
        :param text: The input text to normalize.
        :return: The normalized text.
        """

    @abstractmethod
    def span_normalize(self, text: str) -> List[Tuple[int, int, str]]:
        """
        Normalize the input text and return spans of normalized tokens.
        :param text: The input text to normalize.
        :return: A list of tuples containing the start index, end index, and normalized token for each token span.
        """



"""This module includes Normalizer and NormalizerBuilder"""
from .normalizer import Normalizer
from .tasks.normalizer.multi_lingual_normalizer import MultiLingualNormalizer
from .tasks.normalizer.multi_lingual_normalizer_builder import MultiLingualNormalizerBuilder
from .tasks.normalizer.normalizer_builder import NormalizerBuilder
from .tasks.tokenizer.nltk_tokenizer import NltkTokenizer
from .tasks.tokenizer.spacy_tokenizer import SpacyTokenizer
from .tokenizer import Tokenizer

__all__ = ["Normalizer", "Tokenizer", "NormalizerBuilder", "MultiLingualNormalizer", "MultiLingualNormalizerBuilder",
           "NltkTokenizer", "SpacyTokenizer"]
